
open Graph

module M=Mygraph;;
module MyG=M.VE.G;;
module S=Myset;;
module B=S.Binding;;
module W = struct
  type label = int
  type t = int
  let weight x = x
  let compare = Pervasives.compare
  let add = (+)
  let zero = 0
end;;
module Dijkstra = Path.Dijkstra(MyG)(W);;
module Bfs=Traverse.Bfs(MyG);;

(*
* Module for the transition rules of the automaton.The type 't'
* is a state-an int, and a terminal-a char.
* This module will be used as a key for Map.Make, each key being mapped to
* a list of ints, representing a list of states.
*)
module ST=
  struct
    type t={i:int;c:char}
    let compare = compare
  end;;

module Nonterminal= (Char : Map.OrderedType with type t = char);;
module Terminal = (Char : Map.OrderedType with type t = char);;
module HorsSigma=Map.Make(Terminal);;
module HorsN= Map.Make(Nonterminal);;
module HorsR= Map.Make(Nonterminal);;
module Delta= Map.Make(ST);;
module Dot = Graphviz.Dot(M.VE.Display);;


let handle= open_out "results/example3-1.txt";;

(*The map 'hs' represents the Sigma of the HORS*) 
let hs=ref HorsSigma.empty;;

(*The map 'hn' represents the N of the HORS*) 
let hn=ref HorsN.empty;;

(*The map 'hr' represents the R of the HORS*) 
let hr=ref HorsR.empty;;
let char_arity=ref [];;
(*Used for keeping track of new type variables *)
let alfa=ref 0;;

(*automaton delta*)
let ad=ref Delta.empty;;

(*number of states in the automaton*)
let nrs=ref 0;;

(*Adds a new pair (terminal,kind) to 'hs'*)
let hsadd (c,k)= 
hs:=HorsSigma.add c k (!hs);
char_arity:= List.append !char_arity [(c, Hors.arity k)];;

(*Adds a new pair (non-terminal,kind) to 'hn'*)
let hnadd (c,k)= 
hn:=HorsN.add c k (!hn);
char_arity:= List.append !char_arity [(c, Hors.arity k)];;

(*Adds a new pair (non-terminal,{lambda variables,ground term}) to 'hr'*)
let hradd (c,vv,tt)= 
hr:=HorsR.add c ({Hors.v=vv;Hors.lambdat=tt}) (!hr);;

(*
*Adds a new automaton delta-rule to 'ad'.
*'di' is a state(integer), 'dc' is a terminal
* 'dq' is a list of states (list of integers)
*)
let dadd (di,dc,dq)= 
ad:=Delta.add ({ST.i=di;ST.c=dc}) dq (!ad);;
 
(*
* Implements part3 of the algorithm, for the binding 'elem'. 
* 'elem' is of the form (F, typ). This function  checks if
* the Gamma environment 'envir' proves that R(F) is of type 
* typ from above. If it is, the binding (F, typ) is added 
* to the environment 'envirFinal'. This function will be applied 
* to all the bindings in 'envir'. The ones for which the property
* above holds are added to 'envirFinal'. So, 'envirFinal' will be equal
* to F(Gamma) from the paper.
*)
let part3 elem envir char_arity envirFinal=
let environmentF=ref (!envir) in
let char_arityF=ref (!char_arity) in
let ch=elem.B.c in
let typ=elem.B.ty in
let variables=(HorsR.find ch (!hr)).Hors.v in
let nr=ref 0 in
  S.addVariables variables typ environmentF char_arityF nr ;
let lambdaterm=(HorsR.find ch (!hr)).Hors.lambdat in
let typeF=S.getLastType typ in
  if (S.SetTypes.mem typeF (S.tGamma !environmentF !char_arityF lambdaterm))   
  then 
    envirFinal:=S.SetBindings.add {B.c=ch;B.ty=typ} !envirFinal;;


(* 
* Determines the type of a terminal and adds it to the 
* environment 'envir'.
* 'auttrans' is a pair (state, terminal).
*'l' is a list of states ; 'auttrans' and 'l' together
* form a transition rule of the automaton.
*)
let addterminal auttrans l envir=
let mytype= ref (M.I(0)) in
let len=List.length l in
  if len>0 
  then begin
    let li=List.rev (List.append l [auttrans.ST.i]) in
    mytype:=M.A([M.I(List.nth li 1)],M.I(List.nth li 0)) ;
    for j=2 to len 
    do
      mytype:=M.A([M.I(List.nth li j)],!mytype);
    done
  end
  else 
    mytype:=M.I(auttrans.ST.i);
envir:=S.SetBindings.add {B.c=auttrans.ST.c;B.ty=(!mytype)} !envir;;

(*
* Reads the input file and uses the lexer and parser to 
* structure the data. The input is a HORS and a trivial automaton.
*)
let reading ()  =
try
  let in_channel = open_in "examples/example3-1.txt" in
  let lexbuf = Lexing.from_channel in_channel in
    while true 
    do
      let result = HorsParser.main HorsLexer.token lexbuf in
        (*Hors.output_pair_list handle result;*)
      let resultfinal=Hors.get_pair_list result in
        List.iter hsadd resultfinal;
      let result2 = HorsParser.main2 HorsLexer.token lexbuf in
        (*Hors.output_pair_list handle result2;  *)
      let resultfinal=Hors.get_pair_list result2 in
        List.iter hnadd resultfinal;
      let result3 = HorsParser.main3 HorsLexer.token lexbuf in
        (*Hors.output_triple_list handle result3; *)
      let resultfinal=Hors.get_triple_list result3 in
        List.iter hradd resultfinal; 
      let result4 = HorsParser.autnr HorsLexer.token lexbuf in
        nrs:=result4;
        alfa:=result4+1;
        (*output_string handle (string_of_int result4) ;*)
      let result5 = HorsParser.delta HorsLexer.token lexbuf in 
        (*Automata.output_delta handle result5;*)
      let resultfinal= Automata.get_delta_list result5 in
        List.iter dadd resultfinal;
    done
with HorsLexer.Eof -> output_string handle " "


let main()=
(*Creates an empty graph*)
let g= MyG.create() in
let lgraphnodes=ref [] in

(*list of open nodes*)
let lonodes=ref [] in

(*list of closed nodes*)
let lcnodes=ref [] in 

let first={M.f=Hors.Symbol('S',ref []);M.t=[];M.q=0;M.nnode= 0;M.oc=ref 0} in
let i1 =MyG.V.create (first) in 
  lonodes:=List.append (!lonodes) (first::[]);
  MyG.add_vertex g i1;

(*
* Step 1 of the algorithm in the paper.
* This is where the configuration graph is expanded 50 times.
* If an error node is reached, an error trace is printed to the file 'handle'.
*)
while true 
do

  let taumap=ref M.TauMap.empty in
  let count=ref 0 in
    while ((!count)<50 && ((List.length (!lonodes)) >0))
    do
      let n=List.hd (!lonodes) in
        lonodes:=List.tl (!lonodes);
      let myn=M.first_symbol n.M.f in
        if Char.lowercase myn==myn 
        then
          try
            let lstates=Delta.find {ST.i=n.M.q;ST.c=myn} (!ad) in
              M.expandT g lonodes lcnodes n lstates 
          with Not_found ->
            let i2= 
            MyG.V.create({M.f=n.M.f;M.t=n.M.t;M.q=n.M.q;M.nnode=0;M.oc=ref 0})in
            let (h2,ljj)=(Dijkstra.shortest_path g i1 i2) in
               output_string handle "THE ERROR PATH: ";
            let print_edge ed= let temp_node=(MyG.E.dst ed) in
              output_string handle (""^Hors.string_of_termsb temp_node.M.t^"  ")
            in List.iter print_edge h2;
            exit 0
        else
	  let res= HorsR.find myn (!hr) in
            M.expandN g lonodes lcnodes n (res.Hors.v) (res.Hors.lambdat) ;
          count:=(!count)+1
    done;


(*
*Step 2 of the algorithm.
*Type information is extracted from the graph, 
* for all the prefixes of all terms. The map 'taumap'
* contains the mappings ({prefix,number_of_node}->type).
*)
let attach n= 
  lgraphnodes:=List.append !lgraphnodes [n] in
Bfs.iter attach g;
let lenlgraph=(List.length !lgraphnodes )-1 in
let a=ref (M.I(0)) in
  for i=0 to lenlgraph 
  do
    let n=(List.nth !lgraphnodes i) in
    let lent=(List.length n.M.t)-1 in
    for j=0 to lent 
    do 
      let prefix=ref [n.M.f] in
      for nprefix= 0 to (lent-j) 
      do 
        prefix:=List.append !prefix [(List.nth n.M.t nprefix)];
      done;
      a:= M.assignTypePrefix g n !prefix taumap alfa (ref 0) ;
    done;
    let prefixf=[n.M.f] in
      a:= M.assignTypePrefix g n prefixf taumap alfa (ref 0) ;
  done;


(*
* After applying the function 'f' to 'taumap',
* 'environment' will contain only the bindings for the non-terminals.
* For every binding (F,typ), the function 'elim' has been
* previously applied to typ, and only after added to 'environment'.
*)
let environment=ref S.SetBindings.empty  in
 let f x y= (
   let (c,mytype)=(M.put_bindings x y)  in
   if c!='0' then
   let settypes=S.elim mytype !nrs in
   let func z=
     let tyz=(S.replaceEmptyList(S.removeT z)) in
   environment:= 
     S.SetBindings.add {B.c=c;B.ty=tyz} !environment  in
    S.SetTypes.iter func settypes;
             ) in
M.TauMap.iter f !taumap;


(*Step 3*)

(*
*'typesterminals' is a set that contains the bindings for the
* HORS terminals.
*) 
let typesterminals=ref S.SetBindings.empty in
let h x y =addterminal x y typesterminals in
Delta.iter h !ad;


(*
* 'environmentFinal' is a set containing the bindings for 
* terminals and non-terminals.
*)
let environmentFinal=ref S.SetBindings.empty in
let environment2=ref S.SetBindings.empty  in
environment2:=S.SetBindings.union (!environment) (!typesterminals);


(*
* 'environment' is the Gamma environment from the paper.
*'environmentFinal' will be F(Gamma) after this function is applied.
*)
let f3 elem =part3 elem environment2 char_arity environmentFinal in
S.SetBindings.iter f3 !environment;

(*
* The fixed point computation from Step 3.
* The algorithm is sound, so the while will terminate.
*)
while (S.SetBindings.equal (!environmentFinal) (!environment) ==false ) do
  environment:=!environmentFinal;
  environmentFinal:=S.SetBindings.empty;
  let environment2=ref S.SetBindings.empty  in
  environment2:=S.SetBindings.union (!environment) (!typesterminals);
  let f3 elem =part3 elem environment2 char_arity environmentFinal in
  S.SetBindings.iter f3 !environment;
done;
 
(*
* Checks if the binding S:q0 is in 'environment'.
* If it is, the program exits, as the property is satisfied.
*)
if S.SetBindings.mem {B.c='S';B.ty=M.I(0)} !environment then
begin
  output_string handle "PROPERTY SATISFIED";
  output_char handle '\n';
  S.output_setbindings handle !environment;
  output_char handle '\n';
  exit 0;
end;

(*
*The end of the big while. If the property is not satisfied,
*the program goes back to Step 1.
*)
done



(*Dot.output_graph stdout g*)

let _ = 
  reading() ;
  main();
  close_out handle;
