
open Graph

type atomic_type= 
  I of int
| A of (atomic_type list)*atomic_type;;
 
type literm=Hors.term list;;

(* 
* f is Automata term,the first symbol 
* t is the list of the arguments of f. t is a list of terms
* q is state of automaton
* oc is open 0/close 1 
* nnode is the number of the node-it is used in step 2
*)
type node={f:Hors.term; t:literm; q:int; nnode:int; oc:int ref};;

type node_no_number={fn:Hors.term; tn:literm; qn:int};;

(*Used for numbering the nodes*)
let num_node= ref 0;;

(*Fifo list of open/closed nodes*) 
type ocnodes=(node list)ref;;

(*Fhecks if two types of the form I(i), I(j) are equal*)
let typeints_equal i1 j2 =
match i1 with 
  I(i)-> (match j2 with I(j)-> if i==j then true else false);;


(*Returns the arity of type 'typ'*)
let rec varArity typ=
match typ with 
  I(i)->0
  |A(li,ty)-> 1+ (varArity ty);;

(*
* Returns the list of all intersection types 
* present in 'ty', plus the rightmost type of 'ty'
*)
let rec list_of_type ty=
match ty with 
  I(i)->[[ty]]
  |A(li,myt)->li::(list_of_type myt);;

(*
* If the type 'ty' is of type I(i), then it returns i,
* Otherwise, it returns -5.
*)
let isI ty=
match ty with
  I(i)->i
  |A(li, myt)-> (-5);;

(* checks if type 'ty1' is a subtype of type 'ty2'*)
let rec leq ty1 ty2=
let ok=ref true in
(*let ty2I=isI ty2 in
if ty2I!=(-1) then begin*)
  let list_of_lists1=list_of_type ty1 in
  let list_of_lists2=list_of_type ty2 in
  let len1=List.length list_of_lists1 in
  let len2=List.length list_of_lists2 in
  if len1!=len2 then false 
    else 
    begin
      let i=ref 0 in
      while !i<=(len1-1) do
(*intersection type1 is the i-th intersection of ty1*)
        let interty1=List.nth list_of_lists1 !i  in
        let interty2=List.nth list_of_lists2 !i  in
        let leninterty1=List.length interty1 in
        let leninterty2=List.length interty2 in
        let j=ref 0 in 
        while !j<=(leninterty2-1) do
          let tempok=ref false in
          let elem2= List.nth interty2 !j in
          let elem2I=(isI elem2) in
          if elem2I==(-1) then tempok:=true 
          else begin
          if elem2I>(-5) then
            for k=0 to leninterty1-1 do
               let elem1=List.nth interty1 k in
               let elem1I=(isI elem1) in
               if (elem1I==elem2I) then tempok:=true;
            done
         else  
           for k=0 to leninterty1-1 do
             let elem1=List.nth interty1 k in
             if (leq elem1 elem2) then tempok:=true;
           done;
          end;
          if !tempok==false then begin j:=leninterty2;ok:=false end
          else j:=!j+1;
        done;
        if (!ok==true) then i:=!i+1 
        else i:=len1;
      done;
   !ok;
   end;;
   (*end
else true;;*)



(*checks if the two types 'ty1' and 'ty2' are equal*)
let rec equaltype ty1 ty2 =
let list_of_lists1=list_of_type ty1 in
let list_of_lists2=list_of_type ty2 in
let len1=List.length list_of_lists1 in
let len2=List.length list_of_lists2 in
  if len1!=len2 then false 
  else 
let ok=ref true in
if len1>=2 then 
begin
 for i=0 to len1-2 do
 let li1=List.nth list_of_lists1 i in
  let li2=List.nth list_of_lists2 i in
     let lenli=List.length li1 in
     if lenli!=(List.length li2) then ok:=false 
     else begin
         let liequal=ref [] in
          for j=0 to lenli-1 do
            liequal:=(ref false)::!liequal;
          done;
          for j=0 to lenli-1 do
           for k=0 to lenli-1 do
             if !(List.nth !liequal k)==false then
               (List.nth !liequal k):=equaltype (List.nth li1 j)(List.nth li2 k)
           done
          done;
         for j=0 to lenli-1 do
        ok:=!(List.nth !liequal j) && (!ok)
         done;
   end
  done;
   let l1elem1=List.nth (List.nth list_of_lists1 (len1-1)) 0 in
     let l2elem2=List.nth (List.nth list_of_lists2 (len1-1)) 0 in
  ok:=(typeints_equal l1elem1 l2elem2 )  && (!ok);
  !ok
end
else 
begin
ok:=typeints_equal 
    (List.nth (List.nth list_of_lists1 0) 0) 
    (List.nth (List.nth list_of_lists2 0) 0) ;
!ok;
end;;
 
(*checks if type 'myt' is already in the list 'li'*)
let is_type_in_list myt li=
let len=List.length li in
let ok=ref false in
  for i=0 to len-1 
  do
    if (equaltype myt (List.nth li i)) then ok:=true
  done;
!ok;;

(*represents a term viewed as a list of terms
* -as the first symbol followed by the arguments;
* the number of the node in which the term is found
* also appears in the structure
*)
module TermNode=
  struct
 type t={taut:literm; taun:int}
  let compare n m = 
   let s=(string_of_int n.taun)^(Hors.string_of_terms n.taut) in 
     let r=(string_of_int m.taun)^(Hors.string_of_terms m.taut) in
   Pervasives.compare s r
end;;


module TauMap=Map.Make(TermNode)


module VE=struct
 
(*will be used as a functor for the vertexes of the graph*)
module V = struct
type t= node

let compare n m = 
let s=(Hors.string_of_term n.f)^(Hors.string_of_terms n.t)^(string_of_int n.q) in 
let r=(Hors.string_of_term m.f)^(Hors.string_of_terms m.t)^(string_of_int m.q) in
Pervasives.compare s r

let hash n=n.nnode

let equal n m= 
if n.q!=m.q then false
else 
if (Pervasives.compare (Hors.string_of_term n.f) (Hors.string_of_term m.f)!=0) 
then false 
else if (Pervasives.compare (Hors.string_of_terms n.t) (Hors.string_of_terms m.t)==0) 
     then true
     else false

end

(*will be used as a functor for the edges of the graph*)
module E = struct
type t = int
let compare = compare
let hash=Hashtbl.hash
let equal= (=)
let default =0
end

module G=Imperative.Digraph.ConcreteLabeled(V)(E)

(*using Graphviz*)

(*used for displaying the labels of the nodes in Dot(AT&T) *)
module Display=struct
 let vertex_name v=
    let n= (G.V.label v )in
     (Hors.string_of_term n.f)^Hors.string_of_terms n.t ^
      string_of_int n.q^"_"^(string_of_int !(n.oc))
    
        

let graph_attributes _ =[]
let default_vertex_attributes _ = []
let vertex_attributes _ = []
let default_edge_attributes _ = []
let get_subgraph _= None

let edge_attributes e = match G.E.label e with
t -> [`Style `Bold; `Label (string_of_int t)]

include G
end

module Dot = Graphviz.Dot(Display)
end


(*prints type 'mytype' to 'myfile'*)
let rec output_type myfile mytype=
match mytype with
  I(i)->if i==(-1) then output_string myfile "Top" 
        else output_string myfile ("q"^(string_of_int i)^" ")
|A(li,a)->let len=(List.length li)-2 in
          output_char myfile '(';
          for j=0 to len do
            output_type myfile (List.nth li j);output_char myfile '^';
          done;
          output_type myfile (List.nth li (len+1));
          output_string myfile "->";
          output_type myfile a;
          output_char myfile ')';;

(*
*checks if the terms 't', 'p' have the same structure, 
*without caring about the history of each symbol
*)
let rec eq (t, p)= 
match (t,p) with
  (Hors.App(t1,t2), Hors.App(t3,t4))when (eq (t1, t3))->(eq (t2, t4))
  |(Hors.Symbol(s1,li1),Hors.Symbol(s2,li2)) when (s1==s2) ->true
  |_->false;;


(*checks if the lists of terms 't', 'p' are equal *)
let rec eq_list (t,p)=
match t with  
  te::trest->
       (match p with 
            pe::prest-> eq(te,pe) && eq_list(trest,prest) 
         |[] -> true;)
  |[]-> true;;


(*Returns the first symbol of term 't'*)
let rec first_symbol t=
match t with 
  Hors.App(term1,term2)-> first_symbol term1
  |Hors.Symbol(c,li)-> c;;


(*
* Inserts the node 'son' in the graph 'g'.
* An edge from the node 'parent' to 'son will
* be added, having the label 'nr'.
* 'lopen' and 'lclosed' are the lists of the nodes
* in the graph currently open/closed. 
*)
let insert_node parent g lopen lclosed son nr=
num_node:=!num_node+1;
let add=VE.G.add_edge_e g in
let parentd={f=parent.f;t=parent.t;q=parent.q;nnode=parent.nnode;oc= parent.oc} in
let sonlast={f=son.fn;t=son.tn;q=son.qn;nnode=(!num_node);oc= ref 0} in
let i2= VE.G.V.create (parentd) in
let i1 =VE.G.V.create (sonlast) in
let i1_i2 =VE.G.E.create i2 (nr) i1  in
add i1_i2;
(VE.G.V.label i2).oc:=1;
let s=(first_symbol son.fn) in
   if (Char.lowercase s==s)&&(List.length son.tn==0) 
   then (VE.G.V.label i1).oc:=1
   else lopen:=List.append (!lopen) (sonlast::[]);;


(*Checks if the nodes 'n1', 'n2' are equal.*)
let equal n1 n2=
  if (n1.f==n2.f) && (eq_list (n1.t,n2.t)) && (n1.q==n2.q)
  then true 
  else false;; 


let part_insert (p,n) (a,lb)= 
if (equal p a) 
then (a,List.append lb (n::[]))
else (a,lb);; 


(* Returns the first symbol of 't', including its history *)
let rec first_term t=
match t with 
  Hors.App(term1,term2)-> first_term term1
  |Hors.Symbol(c,li)-> t;;

(* 
* 'k' is a list of terms. If the first symbol in 'k'
* is a non-terminal,say F, the function returns
* (F, 'mytype'). Otherwise, it returns ('0',mytype)
*)
let put_bindings k mytype =
if (List.length k.TermNode.taut)==1 then 
  let ter=List.nth k.TermNode.taut 0 in
  let c=first_symbol ter in
  if Char.uppercase c==c then (c,mytype)
  else ('0',mytype)
else ('0',mytype);;


(* 
* Returns the list of arguments of term 't',
* without the first symbol
*)
let decompose t  =
  let rec loop t l =  
  match t with 
    Hors.App(term1,term2)->  loop term1 (List.append (term2::[]) l) 
    |Hors.Symbol(s3,li)->  l     
in loop t [] ;;
    
(*makes a copy of term 'le'*)    
let  rec copy le= 
match le with 
  Hors.App(a,b)-> Hors.App(copy a,copy b)
  |Hors.Symbol(c,li)->let d=ref [] in d:=!li; Hors.Symbol(c, d);;


(* 
* In the term ter, we have to substitute all occurences 
* of the variable vare with the term le
*)
let rec compose_node (vare,ter) le= 
match ter with 
  Hors.App(a,b)-> 
     Hors.App( (compose_node (vare,a) le),(compose_node (vare,b) le))
  |Hors.Symbol(c,li)->
     if( vare==c) then (copy le) else (copy ter);;


(*
* l is the list of terms with which we have to substitute 
* the variables in the lambda term
*)
let rec compose_nodes (var,ter) l=
match l with  
  le::lrest->
    (match var with 
       vare::varrest-> let newterm=compose_node (vare,ter) le  in 
                       compose_nodes (varrest,newterm) lrest 
       |[] -> ter;)
  |[]-> ter;;

(*
* Used when expanding the configuration graph, Case 1.
* Returns a list of nodes created from the list 'terms'
* and the list 'states'. Every such node is open.
*)
let create_nodes terms states =
let rec loop lterms lstates i=
match lterms with 
  te:: lrest-> 
    (match lstates with 
       qe::lsrest->  
          let tt=decompose te in
          ({fn=(first_term te);tn=tt;qn=qe},i):: (loop lrest lsrest (i+1))
       |[]->[];)
|[]->[]
in loop terms states 1;;

(*
* Expands the node 'n', which starts with a terminal. 
* Used when expanding the configuration graph, case1.
* 'g' is the graph, 'lonodes' and 'lcnodes' are the
* current lists of open/closed nodes, 'lstates' will
* be used when calling the function 'create_nodes' above
*)
let expandT g lonodes lcnodes n lstates  = 
let lterms= n.t  in
let list_nodes=create_nodes lterms lstates  in
if (List.length list_nodes)>0 then
 let finsert_node (son,nr)=insert_node n g lonodes lcnodes son nr  in
 List.iter finsert_node list_nodes;;


(*
* Expands the node 'n', which starts with a non-terminal. 
* Used when expanding the configuration graph, case2.
* 'g' is the graph, 'lonodes' and 'lcnodes' are the
* current lists of open/closed nodes, 'lvars' and 'te' will
* be used when calling the function 'compose_nodes' above
*)
let expandN g lonodes lcnodes n lvars te = 
let lterms = n.t in
let myt = compose_nodes (lvars,te) lterms in
let k=(!num_node +1) in
let newterm=Hors.compute_history myt k in
let newnode={fn=(first_term newterm);tn=(decompose newterm);qn=n.q} in
  insert_node n g lonodes lcnodes newnode 0 ;;
     
(*
* The reachable nodes from the node 'n', in the graph 'g'
* will be added to the list 'l'.
*)               
let rec reachableNodes g n l=
let lsuc=VE.G.succ g n in
let num=(List.length lsuc)-1 in
if (num>=0) then 
begin
  l:=(List.append !l lsuc); 
  for i=0 to num do 
    reachableNodes g (List.nth lsuc i) l;
  done;
end;;  


(*Returns a list of histories for all symbols in 'term'*)
let rec get_list term=
match term with 
  Hors.Symbol(c,l)-> [!l]
  |Hors.App(a,b)-> List.append (get_list a) (get_list b);;


(*Checks if list l1 is a prefix of list l2*)
let is_list_prefix l1 l2=
let len=(List.length l1)-1 in
let bo=ref true in
for i=0 to  len do
  if (List.nth l1 i)!=(List.nth l2 i) then bo:=false;
done;
!bo;;


let are_llists_equal l1 l2=
let ok=ref true in   
if (List.length l1==List.length l2) then 
begin
  let len=(List.length l1)-1 in
  for i=0 to len do
    let one= (List.nth l1 i) in
    let two=(List.nth l2 i) in
    let len1=List.length one in
    if (len1<=List.length two) then 
    begin
      if (len1!=0) then
        if (is_list_prefix one two)==false then ok:=false;
    end
    else ok:=false;
  done;
end
else ok:=false;
!ok;;
 
    

(*
*Checks if the term in node 'n' starts with the term 's0'.
*The history of the symbols is also taken into account.
*)
let starts_with_s0 n s0=
let s1= Hors.string_of_term s0 in 
let t1=(Hors.string_of_term n.f)^(Hors.string_of_terms2 n.t) in
let count=ref 0 in
while (s1.[!count]=='_') 
do
  count:=(!count)+1
done;
let len= ((String.length s1 )-(!count))in
let s=String.sub s1 !count len in
  if String.length s>String.length t1 then (-1)
  else 
    let t=String.sub t1 0 len in
    if Pervasives.compare t s ==0 then 
    begin
      let lis0=get_list s0 in
      let leng=(List.length lis0)-1 in
      let i=ref 0 in
      let j=ref 0 in
      let lit=ref (get_list n.f) in
      while !i<leng 
      do
        let templist=get_list (List.nth n.t !j) in
        lit:=List.append !lit templist;
        j:=!j+1;
        i:=!i+List.length templist;
      done;
      if(are_llists_equal lis0 (!lit)) then  (!j)
      else (-1);
   end
   else (-1);;


(*Checks if the term in the node 'n' is equal to the term 'so'*)
let is_s0 n s0=
let s1=(Hors.string_of_term s0) in 
let t=((Hors.string_of_term n.f)^(Hors.string_of_terms2 n.t)) in  
let count=ref 0 in
while (s1.[!count]=='_') 
do
  count:=(!count)+1
done;
let len= ((String.length s1 )-(!count))in
let s=String.sub s1 !count len in
if Pervasives.compare t s ==0 then 
begin
  let lis0=get_list s0 in
  let leng=(List.length lis0)-1 in
  let lit=ref (get_list n.f) in
  let i=ref 0 in
  let j=ref 0 in
    while !i<leng 
    do
      let templist=get_list (List.nth n.t !j) in
      lit:=List.append !lit templist;
      j:=!j+1;
      i:=!i+List.length templist;
    done;
  if(are_llists_equal lis0 (!lit)) then true
  else false;
end
else false;;

(*Returns the list 'li'. 
* The occurences of character 'c' in string 't'
* are appended to 'li'.
*)
let rec list_of_appearances t c li len_from_start =
try 
  let i= String.index t c in
  li:=List.append (!li) [i+len_from_start];
  let len=(String.length t)-i-1  in
  let s=String.sub t (i+1) len in 
  let len_from_st=(len_from_start+i+1) in list_of_appearances s c li len_from_st
with Not_found->li:=!li;;


(*
* Returns the number of symbols different from '_'
* in the string 's', between the positions 0 and 's'.
*)
let num_characters_before s pos=
let nr=ref 0 in
for i=0 to (pos-1) 
do 
  if s.[i]!='_' then nr:=!nr+1;
done;
!nr;;
 
(*
* Checks if the term in node 'n' contains the term 's0'.
* The history of each symbol is taken into account.
*)
let contains_s0 n s0 =
let ok=ref false in
let s= Hors.string_of_term s0 in 
let t=(Hors.string_of_term n.f)^(Hors.string_of_terms n.t) in
let c=first_symbol s0 in
let lis=get_list s0 in
let lens=String.length s in
let lit=ref (get_list n.f) in
let len=(List.length n.t )-1 in
for i=0 to len 
do
  lit:= List.append (!lit) (get_list (List.nth n.t i)); 
done;
let nrs=String.index s c in 
let news=String.sub s nrs (lens-nrs) in
let list_of_c=ref [] in
list_of_appearances t c list_of_c 0;
let lenlist=(List.length !list_of_c) -1 in
if lenlist >=0 then 
  for i=0 to lenlist 
  do
    let pos=(List.nth !list_of_c i) in
    if String.length t>=(pos+lens-nrs) then
      let newt=String.sub t pos (lens-nrs) in
      if Pervasives.compare news newt==0 then 
      begin
        let num_chars_before= num_characters_before t pos in
        let newlit=ref [] in
        for i=num_chars_before to ((List.length lis)+num_chars_before -1) 
        do
          newlit:=List.append (!newlit) [(List.nth !lit i)];
        done;
        if are_llists_equal lis (!newlit) then ok:=true;  
      end
  done;
!ok;;  
   

(*Returns a term composed of the the list 'li' of terms*)
let term_of_list li=
let t=ref (List.nth li 0) in
let len=(List.length li)-1 in
if len>0 then 
  for i=1 to len 
  do
    t:=Hors.App(!t,List.nth li i)
  done;
!t;;

(*Prints the nodes in the list 'lis' to 'myfile'*)
let rec output_nodes myfile lis=
match lis with
  c::l->output_string myfile ((string_of_int c.nnode)^" "); 
        output_nodes myfile l
  |[]->output_char myfile '\n';;

let is_in_map taumap n t=
try  
  TauMap.find {TermNode.taut=t ;TermNode.taun=n.nnode} !taumap; 
with Not_found->I(-1);;


(*Returns true if 'mytype' is Top, else returns false *)
let value_mytype mytype=
match mytype with
  I(i)-> if i==(-1) then true else false
  |A(a,b)-> false;;


(*
* Extracts type information from the configuration graph.
* The prefix 't' of node 'n' in the graph 'g' is given a type.
* The pair (prefix, type) is introduced in the map 'taumap',
* if it is not already there. 'alfa' ia an integer used for
* numbering the type variables and 'ok' records if the
* following term after the prefix appears in an open node.
*)
let rec assignTypePrefix g n t taumap alfa ok =
let mytype=ref( is_in_map taumap n t) in
if (value_mytype !mytype ) then 
  if (List.length t==((List.length n.t) +1) ) then 
  begin
    taumap:= TauMap.add {TermNode.taut=(n.f)::(n.t) ;TermNode.taun=n.nnode} (I(n.q)) (!taumap);
    mytype:=I(n.q);
  end
  else 
  begin
  let lent=(List.length t)-1 in
  let s0=List.nth n.t lent in
  let li=ref [] in
  let reach=ref [] in
  reachableNodes g n reach;
  let len=(List.length !reach)-1 in
  for i=0 to len 
  do
    let elem=(List.nth !reach i) in
    if (!(elem.oc)==1)  then 
    begin
      let mystart=(starts_with_s0 elem s0) in
      if mystart>(-1) then 
      begin                        
        let newt=(first_term s0)::(decompose s0 ) in
        if (is_s0 n (term_of_list t) ) then 
        begin 
          mytype:=I(n.q); 
          if (is_type_in_list !mytype !li)==false then
              li:=(!mytype)::(!li);     
        end
        else 
        begin
          let type2check=(assignTypePrefix g elem newt taumap  alfa (ref 0) ) in
          if (is_type_in_list type2check !li)==false then
              li:=type2check::(!li);
        end
      end
    end                    
    else 
      if (contains_s0 elem s0 ) then ok:=1;
  done;
  let lastt=List.append t [s0] in
  let ki=(assignTypePrefix g n lastt taumap  alfa (ref 0) ) in
  if (!ok==1)||(!(n.oc)==0) then 
  begin 
    alfa:=!alfa+1; 
    li:=I(!alfa)::(!li);
  end;
  if (List.length !li)==0 then mytype:=A([I(-1)],ki)
  else  mytype:=A(!li,ki);
  ok:=0;
  taumap:= TauMap.add {TermNode.taut=t ;TermNode.taun=n.nnode} (!mytype) (!taumap);
end;
!mytype;;






  






